#include "vamp-hostsdk/PluginHostAdapter.h"
#include <cstdlib>

#include "vamp-hostsdk/Files.h"

namespace vamp
{
  PluginHostAdapter::PluginHostAdapter(const VampPluginDescriptor *descriptor, float inputSampleRate) : plugin(inputSampleRate), m_descriptor(descriptor)
  {
    m_handle = m_descriptor->instantiate(m_descriptor, inputSampleRate);
  }

  PluginHostAdapter::~PluginHostAdapter()
  {
    if (m_handle) m_descriptor->cleanup(m_handle);
  }

  std::vector<std::string> PluginHostAdapter::getPluginPath()
  {
    std::vector<std::string> path;
    std::string envPath;

    if (Files::isNonNative32Bit()) {
      (void)Files::getEnvUtf8("VAMP_PATH_32", envPath);
    } else {
      (void)Files::getEnvUtf8("VAMP_PATH", envPath);
    }

#define DEFAULT_VAMP_PATH "$HOME/vamp:$HOME/.vamp:/usr/local/lib/vamp:/usr/lib/vamp"
#define PATH_SEPARATOR ':'
    if (envPath == "") 
    {
      envPath = DEFAULT_VAMP_PATH;
      std::string home;
      if (Files::getEnvUtf8("HOME", home)) 
      {
        std::string::size_type f;
        while ((f = envPath.find("$HOME")) != std::string::npos && f < envPath.length()) 
        {
          envPath.replace(f, 5, home);
        }
      }
    }

    std::string::size_type index = 0, newindex = 0;

    while ((newindex = envPath.find(PATH_SEPARATOR, index)) < envPath.size()) 
    {
      path.push_back(envPath.substr(index, newindex - index));
      index = newindex + 1;
    }
    
    path.push_back(envPath.substr(index));

    return path;
  }

  bool PluginHostAdapter::initialise(size_t channels, size_t stepSize, size_t blockSize) 
  {
    if (!m_handle) return false;
    return m_descriptor->initialise
        (m_handle,
         (unsigned int)channels,
         (unsigned int)stepSize,
         (unsigned int)blockSize) ?
        true : false;
  }

  void PluginHostAdapter::reset()
  {
    m_descriptor->reset(m_handle);
  }

  PluginHostAdapter::InputDomain PluginHostAdapter::getInputDomain() const
  {
    if (m_descriptor->inputDomain == vampFrequencyDomain) 
    {
      return FrequencyDomain;
    } 
    else 
    {
      return TimeDomain;
    }
  }

  std::string PluginHostAdapter::get_identifier() const
  {
    return m_descriptor->identifier;
  }

  std::string PluginHostAdapter::get_name() const
  {
    return m_descriptor->name;
  }

  std::string PluginHostAdapter::get_description() const
  {
    return m_descriptor->description;
  }

  std::string PluginHostAdapter::get_maker() const
  {
    return m_descriptor->maker;
  }

  PluginHostAdapter::parameter_descriptors PluginHostAdapter::get_parameter_descriptors() const
  {
      parameter_descriptors list;
      for (unsigned int i = 0; i < m_descriptor->parameterCount; ++i) {
          const VampParameterDescriptor *spd = m_descriptor->parameters[i];
          parameter_descriptor pd;
          pd.identifier = spd->identifier;
          pd.name = spd->name;
          pd.description = spd->description;
          pd.unit = spd->unit;
          pd.min_value = spd->minValue;
          pd.max_value = spd->maxValue;
          pd.default_value = spd->defaultValue;
          pd.is_quantized = spd->isQuantized;
          pd.quantize_step = spd->quantizeStep;
          if (pd.is_quantized && spd->valueNames) {
              for (unsigned int j = 0; spd->valueNames[j]; ++j) {
                  pd.value_names.push_back(spd->valueNames[j]);
              }
          }
          list.push_back(pd);
      }
      return list;
  }

  float PluginHostAdapter::get_parameter(std::string param) const
  {
    if (!m_handle) return 0.0;

    for (unsigned int i = 0; i < m_descriptor->parameterCount; ++i) {
        if (param == m_descriptor->parameters[i]->identifier) {
            return m_descriptor->getParameter(m_handle, i);
        }
    }

    return 0.0;
  }

  void PluginHostAdapter::set_parameter(std::string param, float value)
  {
    if (!m_handle) return;

    for (unsigned int i = 0; i < m_descriptor->parameterCount; ++i) {
        if (param == m_descriptor->parameters[i]->identifier) {
            m_descriptor->setParameter(m_handle, i, value);
            return;
        }
    }
  }

  PluginHostAdapter::programs PluginHostAdapter::get_programs() const
  {
      programs list;
      
      for (unsigned int i = 0; i < m_descriptor->programCount; ++i) {
          list.push_back(m_descriptor->programs[i]);
      }
      
      return list;
  }

  std::string PluginHostAdapter::get_current_program() const
  {
      if (!m_handle) return "";

      int pn = m_descriptor->getCurrentProgram(m_handle);
      if (pn < (int)m_descriptor->programCount) {
          return m_descriptor->programs[pn];
      } else {
          return "";
      }
  }

  void PluginHostAdapter::select_program(std::string program)
  {
      if (!m_handle) return;

      for (unsigned int i = 0; i < m_descriptor->programCount; ++i) {
          if (program == m_descriptor->programs[i]) {
              m_descriptor->selectProgram(m_handle, i);
              return;
          }
      }
  }

  size_t PluginHostAdapter::getPreferredStepSize() const
  {
      if (!m_handle) return 0;
      return m_descriptor->getPreferredStepSize(m_handle);
  }

  size_t PluginHostAdapter::getPreferredBlockSize() const
  {
      if (!m_handle) return 0;
      return m_descriptor->getPreferredBlockSize(m_handle);
  }

  size_t PluginHostAdapter::getMinChannelCount() const
  {
      if (!m_handle) return 0;
      return m_descriptor->getMinChannelCount(m_handle);
  }

  size_t
  PluginHostAdapter::getMaxChannelCount() const
  {
      if (!m_handle) return 0;
      return m_descriptor->getMaxChannelCount(m_handle);
  }

  PluginHostAdapter::OutputList
  PluginHostAdapter::getOutputDescriptors() const
  {
      OutputList list;
      if (!m_handle) {
  //        std::cerr << "PluginHostAdapter::getOutputDescriptors: no handle " << std::endl;
          return list;
      }

      unsigned int count = m_descriptor->getOutputCount(m_handle);

      for (unsigned int i = 0; i < count; ++i) {
          VampOutputDescriptor *sd = m_descriptor->getOutputDescriptor(m_handle, i);
          OutputDescriptor d;
          d.identifier = sd->identifier;
          d.name = sd->name;
          d.description = sd->description;
          d.unit = sd->unit;
          d.hasFixedBinCount = sd->hasFixedBinCount;
          d.binCount = sd->binCount;
          if (d.hasFixedBinCount && sd->binNames) {
              for (unsigned int j = 0; j < sd->binCount; ++j) {
                  d.binNames.push_back(sd->binNames[j] ? sd->binNames[j] : "");
              }
          }
          d.hasKnownExtents = sd->hasKnownExtents;
          d.minValue = sd->minValue;
          d.maxValue = sd->maxValue;
          d.isQuantized = sd->isQuantized;
          d.quantizeStep = sd->quantizeStep;

          switch (sd->sampleType) {
          case vampOneSamplePerStep:
              d.sampleType = OutputDescriptor::OneSamplePerStep; break;
          case vampFixedSampleRate:
              d.sampleType = OutputDescriptor::FixedSampleRate; break;
          case vampVariableSampleRate:
              d.sampleType = OutputDescriptor::VariableSampleRate; break;
          }

          d.sampleRate = sd->sampleRate;

          if (m_descriptor->vampApiVersion >= 2) {
              d.hasDuration = sd->hasDuration;
          } else {
              d.hasDuration = false;
          }

          list.push_back(d);

          m_descriptor->releaseOutputDescriptor(sd);
      }

      return list;
  }

  PluginHostAdapter::FeatureSet
  PluginHostAdapter::process(const float *const *inputBuffers,
                             real_time timestamp)
  {
      FeatureSet fs;
      if (!m_handle) return fs;

      int sec = timestamp.sec;
      int nsec = timestamp.nsec;
      
      VampFeatureList *features = m_descriptor->process(m_handle,
                                                        inputBuffers,
                                                        sec, nsec);
      
      convertFeatures(features, fs);
      m_descriptor->releaseFeatureSet(features);
      return fs;
  }

  PluginHostAdapter::FeatureSet
  PluginHostAdapter::getRemainingFeatures()
  {
      FeatureSet fs;
      if (!m_handle) return fs;
      
      VampFeatureList *features = m_descriptor->getRemainingFeatures(m_handle); 

      convertFeatures(features, fs);
      m_descriptor->releaseFeatureSet(features);
      return fs;
  }

  void
  PluginHostAdapter::convertFeatures(VampFeatureList *features,
                                     FeatureSet &fs)
  {
      if (!features) return;

      unsigned int outputs = m_descriptor->getOutputCount(m_handle);

      for (unsigned int i = 0; i < outputs; ++i) {
          
          VampFeatureList &list = features[i];

          if (list.featureCount > 0) {

              Feature feature;
              feature.values.reserve(list.features[0].v1.valueCount);

              for (unsigned int j = 0; j < list.featureCount; ++j) {

                  feature.hasTimestamp = list.features[j].v1.hasTimestamp;
                  feature.timestamp = real_time(list.features[j].v1.sec,
                                               list.features[j].v1.nsec);
                  feature.hasDuration = false;

                  if (m_descriptor->vampApiVersion >= 2) {
                      unsigned int j2 = j + list.featureCount;
                      feature.hasDuration = list.features[j2].v2.hasDuration;
                      feature.duration = real_time(list.features[j2].v2.durationSec,
                                                  list.features[j2].v2.durationNsec);
                  }

                  for (unsigned int k = 0; k < list.features[j].v1.valueCount; ++k) {
                      feature.values.push_back(list.features[j].v1.values[k]);
                  }

                  if (list.features[j].v1.label) {
                      feature.label = list.features[j].v1.label;
                  }

                  fs[i].push_back(feature);

                  if (list.features[j].v1.valueCount > 0) {
                      feature.values.clear();
                  }

                  if (list.features[j].v1.label) {
                      feature.label = "";
                  }
              }
          }
      }
  }
}
