#include "vamp-hostsdk/PluginSummarisingAdapter.h"

#include <cmath>
#include <algorithm>
#include <climits>

namespace vamp::host 
{
  class PluginSummarisingAdapter::Impl
  {
    public:
      Impl(plugin* p, float inputSampleRate);
      ~Impl();

      bool initialise(size_t channels, size_t stepSize, size_t blockSize);

      void reset();

      FeatureSet process(const float *const *inputBuffers, real_time timestamp);
      FeatureSet getRemainingFeatures();

      void setSummarySegmentBoundaries(const SegmentBoundaries &);

      FeatureList getSummaryForOutput(int output,
                                      SummaryType type,
                                      AveragingMethod avg);

      FeatureSet getSummaryForAllOutputs(SummaryType type,
                                         AveragingMethod avg);

    protected:
      plugin *m_plugin;
      float m_inputSampleRate;
      size_t m_stepSize;
      size_t m_blockSize;

      SegmentBoundaries m_boundaries;

      typedef std::vector<float> ValueList;

      struct Result { // smaller than Feature
          real_time time;
          real_time duration;
          ValueList values; // bin number -> value
      };

      typedef std::vector<Result> ResultList;

      struct OutputAccumulator {
          int bins;
          ResultList results;
          OutputAccumulator() : bins(0) { }
      };

      typedef std::map<int, OutputAccumulator> OutputAccumulatorMap;
      OutputAccumulatorMap m_accumulators; // output number -> accumulator

      typedef std::map<real_time, OutputAccumulator> SegmentAccumulatorMap;
      typedef std::map<int, SegmentAccumulatorMap> OutputSegmentAccumulatorMap;
      OutputSegmentAccumulatorMap m_segmentedAccumulators; // output -> segmented

      typedef std::map<int, real_time> OutputTimestampMap;
      OutputTimestampMap m_prevTimestamps; // output number -> timestamp
      OutputTimestampMap m_prevDurations; // output number -> durations

      struct OutputBinSummary {

          int count;

          // extents
          double minimum;
          double maximum;
          double sum;

          // sample-average results
          double median;
          double mode;
          double variance;

          // continuous-time average results
          double median_c;
          double mode_c;
          double mean_c;
          double variance_c;
      };

      typedef std::map<int, OutputBinSummary> OutputSummary;
      typedef std::map<real_time, OutputSummary> SummarySegmentMap;
      typedef std::map<int, SummarySegmentMap> OutputSummarySegmentMap;

      OutputSummarySegmentMap m_summaries;

      bool m_reduced;
      real_time m_endTime;

      void accumulate(const FeatureSet &fs, real_time, bool final);
      void accumulate(int output, const Feature &f, real_time, bool final);
      void accumulateFinalDurations();
      void findSegmentBounds(real_time t, real_time &start, real_time &end);
      void segment();
      void reduce();

      std::string getSummaryLabel(SummaryType type, AveragingMethod avg);
  };

  static real_time INVALID_DURATION(INT_MIN, INT_MIN);
      
  PluginSummarisingAdapter::PluginSummarisingAdapter(plugin* p) : PluginWrapper(p)
  {
    m_impl = new Impl(p, m_inputSampleRate);
  }

  PluginSummarisingAdapter::~PluginSummarisingAdapter()
  {
    delete m_impl;
  }

  bool
  PluginSummarisingAdapter::initialise(size_t channels,
                                       size_t stepSize, size_t blockSize)
  {
      return
          PluginWrapper::initialise(channels, stepSize, blockSize) &&
          m_impl->initialise(channels, stepSize, blockSize);
  }

  void
  PluginSummarisingAdapter::reset()
  {
      m_impl->reset();
  }

  plugin::FeatureSet
  PluginSummarisingAdapter::process(const float *const *inputBuffers, real_time timestamp)
  {
    return m_impl->process(inputBuffers, timestamp);
  }

  plugin::FeatureSet
  PluginSummarisingAdapter::getRemainingFeatures()
  {
    return m_impl->getRemainingFeatures();
  }

  void
  PluginSummarisingAdapter::setSummarySegmentBoundaries(const SegmentBoundaries &b)
  {
    m_impl->setSummarySegmentBoundaries(b);
  }

  plugin::FeatureList
  PluginSummarisingAdapter::getSummaryForOutput(int output, SummaryType type, AveragingMethod avg) 
  {
    return m_impl->getSummaryForOutput(output, type, avg);
  }

  plugin::FeatureSet PluginSummarisingAdapter::getSummaryForAllOutputs(SummaryType type, AveragingMethod avg)
  {
    return m_impl->getSummaryForAllOutputs(type, avg);
  }

  PluginSummarisingAdapter::Impl::Impl(plugin* p, float inputSampleRate) : m_plugin(p), m_inputSampleRate(inputSampleRate), m_reduced(false)
  {
  }

  PluginSummarisingAdapter::Impl::~Impl()
  {
  }

  bool PluginSummarisingAdapter::Impl::initialise(size_t, size_t stepSize, size_t blockSize)
  {
    m_stepSize = stepSize;
    m_blockSize = blockSize;
    return true;
  }

  void PluginSummarisingAdapter::Impl::reset()
  {
    m_accumulators.clear();
    m_segmentedAccumulators.clear();
    m_prevTimestamps.clear();
    m_prevDurations.clear();
    m_summaries.clear();
    m_reduced = false;
    m_endTime = real_time();
    m_plugin->reset();
  }

  plugin::FeatureSet PluginSummarisingAdapter::Impl::process(const float *const *inputBuffers, real_time timestamp)
  {
      if (m_reduced) {
        std::cerr << "WARNING: Cannot call PluginSummarisingAdapter::process() or getRemainingFeatures() after one of the getSummary methods" << std::endl;
      }
      FeatureSet fs = m_plugin->process(inputBuffers, timestamp);
      accumulate(fs, timestamp, false);
      m_endTime = timestamp + real_time::f2rt(m_stepSize, int(m_inputSampleRate + 0.5));
      return fs;
  }

  plugin::FeatureSet PluginSummarisingAdapter::Impl::getRemainingFeatures()
  {
      if (m_reduced) {
        std::cerr << "WARNING: Cannot call PluginSummarisingAdapter::process() or getRemainingFeatures() after one of the getSummary methods" << std::endl;
      }
      FeatureSet fs = m_plugin->getRemainingFeatures();
      accumulate(fs, m_endTime, true);
      return fs;
  }

  void
  PluginSummarisingAdapter::Impl::setSummarySegmentBoundaries(const SegmentBoundaries &b)
  {
      m_boundaries = b;
#ifdef DEBUG_PLUGIN_SUMMARISING_ADAPTER
      cerr << "PluginSummarisingAdapter::setSummarySegmentBoundaries: boundaries are:" << endl;
      for (SegmentBoundaries::const_iterator i = m_boundaries.begin();
           i != m_boundaries.end(); ++i) {
          cerr << *i << "  ";
      }
      cerr << endl;
#endif
  }

  plugin::FeatureList
  PluginSummarisingAdapter::Impl::getSummaryForOutput(int output,
                                                      SummaryType type,
                                                      AveragingMethod avg)
  {
      if (!m_reduced) {
          accumulateFinalDurations();
          segment();
          reduce();
          m_reduced = true;
      }

      bool continuous = (avg == ContinuousTimeAverage);

      FeatureList fl;
      for (SummarySegmentMap::const_iterator i = m_summaries[output].begin();
           i != m_summaries[output].end(); ++i) {

          Feature f;

          f.hasTimestamp = true;
          f.timestamp = i->first;

          f.hasDuration = true;
          SummarySegmentMap::const_iterator ii = i;
          if (++ii == m_summaries[output].end()) {
              f.duration = m_endTime - f.timestamp;
          } else {
              f.duration = ii->first - f.timestamp;
          }

          f.label = getSummaryLabel(type, avg);

          for (OutputSummary::const_iterator j = i->second.begin();
               j != i->second.end(); ++j) {

              // these will be ordered by bin number, and no bin numbers
              // will be missing except at the end (because of the way
              // the accumulators were initially filled in accumulate())

              const OutputBinSummary &summary = j->second;
              double result = 0.f;

              switch (type) {

              case Minimum:
                  result = summary.minimum;
                  break;

              case Maximum:
                  result = summary.maximum;
                  break;

              case Mean:
                  if (continuous) {
                      result = summary.mean_c;
                  } else if (summary.count) {
                      result = summary.sum / summary.count;
                  }
                  break;

              case Median:
                  if (continuous) result = summary.median_c;
                  else result = summary.median;
                  break;

              case Mode:
                  if (continuous) result = summary.mode_c;
                  else result = summary.mode;
                  break;

              case Sum:
                  result = summary.sum;
                  break;

              case Variance:
                  if (continuous) result = summary.variance_c;
                  else result = summary.variance;
                  break;

              case StandardDeviation:
                  if (continuous) result = sqrt(summary.variance_c);
                  else result = sqrt(summary.variance);
                  break;

              case Count:
                  result = summary.count;
                  break;

              case UnknownSummaryType:
                  break;

              default:
                  break;
              }
              
              f.values.push_back(float(result));
          }

          fl.push_back(f);
      }
      return fl;
  }

  plugin::FeatureSet
  PluginSummarisingAdapter::Impl::getSummaryForAllOutputs(SummaryType type,
                                                          AveragingMethod avg)
  {
      if (!m_reduced) {
          accumulateFinalDurations();
          segment();
          reduce();
          m_reduced = true;
      }

      FeatureSet fs;
      for (OutputSummarySegmentMap::const_iterator i = m_summaries.begin();
           i != m_summaries.end(); ++i) {
          fs[i->first] = getSummaryForOutput(i->first, type, avg);
      }
      return fs;
  }

  void
  PluginSummarisingAdapter::Impl::accumulate(const FeatureSet &fs,
                                             real_time timestamp, 
                                             bool final)
  {
      for (FeatureSet::const_iterator i = fs.begin(); i != fs.end(); ++i) {
          for (FeatureList::const_iterator j = i->second.begin();
               j != i->second.end(); ++j) {
              if (j->hasTimestamp) {
                  accumulate(i->first, *j, j->timestamp, final);
              } else {
                  //!!! is this correct?
                  accumulate(i->first, *j, timestamp, final);
              }
          }
      }
  }

  std::string
  PluginSummarisingAdapter::Impl::getSummaryLabel(SummaryType type,
                                                  AveragingMethod avg)
  {
    std::string label;
    std::string avglabel;

      if (avg == SampleAverage) avglabel = ", sample average";
      else avglabel = ", continuous-time average";

      switch (type) {
      case Minimum:  label = "(minimum value)"; break;
      case Maximum:  label = "(maximum value)"; break;
      case Mean:     label = "(mean value" + avglabel + ")"; break;
      case Median:   label = "(median value" + avglabel + ")"; break;
      case Mode:     label = "(modal value" + avglabel + ")"; break;
      case Sum:      label = "(sum)"; break;
      case Variance: label = "(variance" + avglabel + ")"; break;
      case StandardDeviation: label = "(standard deviation" + avglabel + ")"; break;
      case Count:    label = "(count)"; break;
      case UnknownSummaryType: label = "(unknown summary)"; break;
      }
      
      return label;
  }

  void
  PluginSummarisingAdapter::Impl::accumulate(int output, const Feature &f, real_time timestamp, bool)
  {
      if (m_prevDurations.find(output) != m_prevDurations.end()) {

          real_time prevDuration;

          // Note that m_prevDurations[output] only contains the
          // duration field that was contained in the previous feature.
          // If it didn't have an explicit duration,
          // m_prevDurations[output] should be INVALID_DURATION and we
          // will have to calculate the duration from the previous and
          // current timestamps.

          if (m_prevDurations[output] != INVALID_DURATION) {
              prevDuration = m_prevDurations[output];
          } else {
              prevDuration = timestamp - m_prevTimestamps[output];
          }

          
          m_accumulators[output].results
              [m_accumulators[output].results.size() - 1]
              .duration = prevDuration;
      }

      if (f.hasDuration) m_prevDurations[output] = f.duration;
      else m_prevDurations[output] = INVALID_DURATION;

      m_prevTimestamps[output] = timestamp;

      if (f.hasDuration) {
          real_time et = timestamp;
          et = et + f.duration;
          if (et > m_endTime) m_endTime = et;
      }

      Result result;
      result.time = timestamp;
      result.duration = INVALID_DURATION;

      if (int(f.values.size()) > m_accumulators[output].bins) {
          m_accumulators[output].bins = int(f.values.size());
      }

      for (int i = 0; i < int(f.values.size()); ++i) {
          result.values.push_back(f.values[i]);
      }

      m_accumulators[output].results.push_back(result);
  }

  void
  PluginSummarisingAdapter::Impl::accumulateFinalDurations()
  {
      for (OutputTimestampMap::iterator i = m_prevTimestamps.begin();
           i != m_prevTimestamps.end(); ++i) {

          int output = i->first;

          int acount = int(m_accumulators[output].results.size());

          if (acount == 0) continue;

          if (m_prevDurations.find(output) != m_prevDurations.end() &&
              m_prevDurations[output] != INVALID_DURATION) {
              m_accumulators[output].results[acount - 1].duration =
                  m_prevDurations[output];

          } else {

              m_accumulators[output].results[acount - 1].duration =
                  m_endTime - m_prevTimestamps[output];
          }
          
      }
  }

  void
  PluginSummarisingAdapter::Impl::findSegmentBounds(real_time t,
                                                    real_time &start,
                                                    real_time &end)
  {
#ifdef DEBUG_PLUGIN_SUMMARISING_ADAPTER_SEGMENT
      cerr << "findSegmentBounds: t = " << t <<  endl;
#endif

      SegmentBoundaries::const_iterator i = std::upper_bound(m_boundaries.begin(), m_boundaries.end(), t);

      start = real_time::zero_time;
      end = m_endTime;

      if (i != m_boundaries.end()) {
          end = *i;
      }

      if (i != m_boundaries.begin()) {
          start = *--i;
      }
  }

  void
  PluginSummarisingAdapter::Impl::segment()
  {

      for (OutputAccumulatorMap::iterator i = m_accumulators.begin();
           i != m_accumulators.end(); ++i) {

          int output = i->first;
          OutputAccumulator &source = i->second;

          for (int n = 0; n < int(source.results.size()); ++n) {
              real_time resultStart = source.results[n].time;
              real_time resultEnd = resultStart + source.results[n].duration;

              real_time segmentStart = real_time::zero_time;
              real_time segmentEnd = resultEnd - real_time(1, 0);
              
              real_time prevSegmentStart = segmentStart - real_time(1, 0);

              while (segmentEnd < resultEnd) {

                  findSegmentBounds(resultStart, segmentStart, segmentEnd);

                  if (segmentStart == prevSegmentStart) {
                      break;
                  }
                  prevSegmentStart = segmentStart;
                  
                  real_time chunkStart = resultStart;
                  if (chunkStart < segmentStart) chunkStart = segmentStart;

                  real_time chunkEnd = resultEnd;
                  if (chunkEnd > segmentEnd) chunkEnd = segmentEnd;
                  
                  m_segmentedAccumulators[output][segmentStart].bins = source.bins;

                  Result chunk;
                  chunk.time = chunkStart;
                  chunk.duration = chunkEnd - chunkStart;
                  chunk.values = source.results[n].values;

                  m_segmentedAccumulators[output][segmentStart].results.push_back(chunk);

                  resultStart = chunkEnd;
              }
          }
      }
  }

  struct ValueDurationFloatPair
  {
      float value;
      float duration;

      ValueDurationFloatPair() : value(0), duration(0) { }
      ValueDurationFloatPair(float v, float d) : value(v), duration(d) { }
      ValueDurationFloatPair &operator=(const ValueDurationFloatPair &p) {
          value = p.value;
          duration = p.duration;
          return *this;
      }
      bool operator<(const ValueDurationFloatPair &p) const {
          return value < p.value;
      }
  };

  static double toSec(const real_time &r)
  {
      return r.sec + double(r.nsec) / 1000000000.0;
  }

  void
  PluginSummarisingAdapter::Impl::reduce()
  {
      for (OutputSegmentAccumulatorMap::iterator i =
               m_segmentedAccumulators.begin();
           i != m_segmentedAccumulators.end(); ++i) {

          int output = i->first;
          SegmentAccumulatorMap &segments = i->second;

          for (SegmentAccumulatorMap::iterator j = segments.begin();
               j != segments.end(); ++j) {

              real_time segmentStart = j->first;
              OutputAccumulator &accumulator = j->second;

              int sz = int(accumulator.results.size());

#ifdef DEBUG_PLUGIN_SUMMARISING_ADAPTER
              cerr << "reduce: segment starting at " << segmentStart
                        << " on output " << output << " has " << sz << " result(s)" << endl;
#endif

              double totalDuration = 0.0;
              //!!! is this right?
              if (sz > 0) {
#ifdef DEBUG_PLUGIN_SUMMARISING_ADAPTER
                  cerr << "last time = " << accumulator.results[sz-1].time 
                            << ", duration = " << accumulator.results[sz-1].duration
                            << " (step = " << m_stepSize << ", block = " << m_blockSize << ")"
                            << endl;
#endif
                  totalDuration = toSec((accumulator.results[sz-1].time +
                                         accumulator.results[sz-1].duration) -
                                        segmentStart);
              }

              for (int bin = 0; bin < accumulator.bins; ++bin) {

                  // work on all values over time for a single bin

#ifdef DEBUG_PLUGIN_SUMMARISING_ADAPTER
                  cerr << "bin " << bin << ":" << endl;
#endif
                  
                  OutputBinSummary summary;

                  summary.count = sz;

                  summary.minimum = 0.f;
                  summary.maximum = 0.f;

                  summary.median = 0.f;
                  summary.mode = 0.f;
                  summary.sum = 0.f;
                  summary.variance = 0.f;

                  summary.median_c = 0.f;
                  summary.mode_c = 0.f;
                  summary.mean_c = 0.f;
                  summary.variance_c = 0.f;

                  if (sz == 0) continue;

                  std::vector<ValueDurationFloatPair> valvec;

                  for (int k = 0; k < sz; ++k) {
                      while (int(accumulator.results[k].values.size()) <
                             accumulator.bins) {
                          accumulator.results[k].values.push_back(0.f);
                      }
                  }

                  for (int k = 0; k < sz; ++k) {
                      float value = accumulator.results[k].values[bin];
                      valvec.push_back
                          (ValueDurationFloatPair
                           (value,
                            float(toSec(accumulator.results[k].duration))));
                  }

                  sort(valvec.begin(), valvec.end());

                  summary.minimum = valvec[0].value;
                  summary.maximum = valvec[sz-1].value;

#ifdef DEBUG_PLUGIN_SUMMARISING_ADAPTER
                  cerr << "total duration = " << totalDuration << endl;
#endif

#ifdef DEBUG_PLUGIN_SUMMARISING_ADAPTER
  /*
                  cerr << "value vector for medians:" << endl;
                  for (int k = 0; k < sz; ++k) {
                      cerr << "(" << valvec[k].value << "," << valvec[k].duration << ") ";
                  }
                  cerr << endl;
  */
#endif

                  if (sz % 2 == 1) {
                      summary.median = valvec[sz/2].value;
                  } else {
                      summary.median = (valvec[sz/2].value + valvec[sz/2 + 1].value) / 2;
                  }
              
                  double duracc = 0.0;
                  summary.median_c = valvec[sz-1].value;

                  for (int k = 0; k < sz; ++k) {
                      duracc += valvec[k].duration;
                      if (duracc > totalDuration/2) {
                          summary.median_c = valvec[k].value;
                          break;
                      }
                  }

#ifdef DEBUG_PLUGIN_SUMMARISING_ADAPTER
                  cerr << "median_c = " << summary.median_c << endl;
                  cerr << "median = " << summary.median << endl;
#endif
                  
                  std::map<float, int> distribution;

#ifdef DEBUG_PLUGIN_SUMMARISING_ADAPTER
                  cerr << "summing (discrete): ";
#endif
                  for (int k = 0; k < sz; ++k) {
#ifdef DEBUG_PLUGIN_SUMMARISING_ADAPTER
                      cerr << accumulator.results[k].values[bin] << " ";
#endif
                      summary.sum += accumulator.results[k].values[bin];
                      distribution[accumulator.results[k].values[bin]] += 1;
                  }
#ifdef DEBUG_PLUGIN_SUMMARISING_ADAPTER
                  cerr << endl;
#endif

                  int md = 0;

                  for (std::map<float, int>::iterator di = distribution.begin();
                       di != distribution.end(); ++di) {
                      if (di->second > md) {
                          md = di->second;
                          summary.mode = di->first;
                      }
                  }

                  distribution.clear();

                  std::map<float, double> distribution_c;

                  for (int k = 0; k < sz; ++k) {
                      distribution_c[accumulator.results[k].values[bin]]
                          += toSec(accumulator.results[k].duration);
                  }

                  double mrd = 0.0;

                  for (std::map<float, double>::iterator di = distribution_c.begin();
                       di != distribution_c.end(); ++di) {
                      if (di->second > mrd) {
#ifdef DEBUG_PLUGIN_SUMMARISING_ADAPTER
                          cerr << "element " << di->first << " spans time "
                               << di->second << " and so is an improved mode "
                               << "candidate over element " << summary.mode_c
                               << " which spanned " << mrd << endl;
#endif
                          mrd = di->second;
                          summary.mode_c = di->first;
                      }
                  }

                  distribution_c.clear();

                  if (totalDuration > 0.0) {

                      double sum_c = 0.0;

#ifdef DEBUG_PLUGIN_SUMMARISING_ADAPTER
                      cerr << "summing (continuous): ";
#endif
                      for (int k = 0; k < sz; ++k) {
#ifdef DEBUG_PLUGIN_SUMMARISING_ADAPTER
                          cerr << accumulator.results[k].values[bin] << "*"
                               << toSec(accumulator.results[k].duration) << " ";
#endif
                          double value = accumulator.results[k].values[bin]
                              * toSec(accumulator.results[k].duration);
                          sum_c += value;
                      }
#ifdef DEBUG_PLUGIN_SUMMARISING_ADAPTER
                  cerr << endl;
#endif

#ifdef DEBUG_PLUGIN_SUMMARISING_ADAPTER
                      cerr << "mean_c = " << sum_c << " / " << totalDuration << " = "
                                << sum_c / totalDuration << " (sz = " << sz << ")" << endl;
#endif
                  
                      summary.mean_c = sum_c / totalDuration;

                      for (int k = 0; k < sz; ++k) {
                          double value = accumulator.results[k].values[bin];
  //                            * toSec(accumulator.results[k].duration);
                          summary.variance_c +=
                              (value - summary.mean_c) * (value - summary.mean_c)
                              * toSec(accumulator.results[k].duration);
                      }

  //                    summary.variance_c /= summary.count;
                      summary.variance_c /= totalDuration;
                  }

                  double mean = summary.sum / summary.count;

#ifdef DEBUG_PLUGIN_SUMMARISING_ADAPTER
                  cerr << "mean = " << summary.sum << " / " << summary.count << " = "
                            << summary.sum / summary.count << endl;
#endif

                  for (int k = 0; k < sz; ++k) {
                      float value = accumulator.results[k].values[bin];
                      summary.variance += (value - mean) * (value - mean);
                  }
                  summary.variance /= summary.count;

                  m_summaries[output][segmentStart][bin] = summary;
              }
          }
      }

      m_segmentedAccumulators.clear();
      m_accumulators.clear();
  }
}
