#pragma once

#include <string>
#include <vector>
#include <map>

#include "real_time/real_time.hpp"

namespace vamp 
{
  class plugin
  {
    public:
      virtual ~plugin() { }

      enum InputDomain { TimeDomain, FrequencyDomain };

      struct OutputDescriptor
      {
        std::string identifier;
        std::string name;
        std::string description;
        std::string unit;
        bool hasFixedBinCount;
        size_t binCount;
        std::vector<std::string> binNames;
        bool hasKnownExtents;
        float minValue;
        float maxValue;
        bool isQuantized;
        float quantizeStep;

        enum SampleType 
        {
          OneSamplePerStep,
          FixedSampleRate,
          VariableSampleRate
        };

        SampleType sampleType;
        float sampleRate;

        bool hasDuration;

        OutputDescriptor() :
          hasFixedBinCount(false),
          binCount(0),
          hasKnownExtents(false),
          minValue(0),
          maxValue(0),
          isQuantized(false),
          quantizeStep(0),
          sampleType(OneSamplePerStep), 
          sampleRate(0), 
          hasDuration(false) { }
      };

      using OutputList = std::vector<OutputDescriptor>;

      struct Feature
      {
        bool hasTimestamp;
        real_time timestamp;
        bool hasDuration;
        real_time duration;
        std::vector<float> values;
        std::string label;
        Feature() : hasTimestamp(false), hasDuration(false) { }
      };

      using FeatureList = std::vector<Feature>;
      using FeatureSet = std::map<int, FeatureList>; 

      virtual OutputList getOutputDescriptors() const = 0;
      virtual bool initialise(size_t inputChannels, size_t stepSize, size_t blockSize) = 0;
      virtual void reset() = 0;
      virtual InputDomain getInputDomain() const = 0;
      virtual size_t getPreferredBlockSize() const { return 0; }
      virtual size_t getPreferredStepSize() const { return 0; }
      virtual size_t getMinChannelCount() const { return 1; }
      virtual size_t getMaxChannelCount() const { return 1; }
      virtual FeatureSet process(const float *const *inputBuffers, real_time timestamp) = 0;
      virtual FeatureSet getRemainingFeatures() = 0;
      virtual std::string get_type() const { return "Feature Extraction Plugin"; }

      float getInputSampleRate() const { return m_inputSampleRate; }


      struct parameter_descriptor
      {
        std::string identifier;
        std::string name;
        std::string description;
        std::string unit;
        float min_value { 0.f };
        float max_value { 0.f };
        float default_value { 0.f };
        bool is_quantized { false };
        float quantize_step { 0.f };
        std::vector<std::string> value_names;
      };

      using parameter_descriptors = std::vector<parameter_descriptor>;
      using programs = std::vector<std::string>;

    public:
      virtual std::string get_identifier() const = 0;
      virtual std::string get_name() const = 0;
      virtual std::string get_description() const = 0;
      virtual std::string get_maker() const = 0;
      virtual float get_parameter(std::string) const { return 0.f; }
      virtual std::string get_current_program() const { return ""; }
      virtual parameter_descriptors get_parameter_descriptors() const { return parameter_descriptors(); }
      virtual programs get_programs() const { return programs(); }

      virtual void set_parameter(std::string, float) { } 
      virtual void select_program(std::string) { }
      
    protected:
      plugin(float inputSampleRate) : m_inputSampleRate(inputSampleRate) { }
      float m_inputSampleRate;
  };
}
