#include "vamp-sdk/PluginAdapter.h"

#include <cstring>
#include <cstdlib>

#include <mutex>

namespace vamp
{
  class PluginAdapterBase::Impl
  {
    public:
      Impl(PluginAdapterBase *);
      ~Impl();

      const VampPluginDescriptor *getDescriptor();

    protected:
      PluginAdapterBase *m_base;

      static VampPluginHandle vampInstantiate(const VampPluginDescriptor *desc,
                                              float inputSampleRate);

      static void vampCleanup(VampPluginHandle handle);

      static int vampInitialise(VampPluginHandle handle, unsigned int channels,
                               unsigned int stepSize, unsigned int blockSize);

      static void vampReset(VampPluginHandle handle);

      static float vampGetParameter(VampPluginHandle handle, int param);
      static void vampSetParameter(VampPluginHandle handle, int param, float value);

      static unsigned int vampGetCurrentProgram(VampPluginHandle handle);
      static void vampSelectProgram(VampPluginHandle handle, unsigned int program);

      static unsigned int vampGetPreferredStepSize(VampPluginHandle handle);
      static unsigned int vampGetPreferredBlockSize(VampPluginHandle handle);
      static unsigned int vampGetMinChannelCount(VampPluginHandle handle);
      static unsigned int vampGetMaxChannelCount(VampPluginHandle handle);

      static unsigned int vampGetOutputCount(VampPluginHandle handle);

      static VampOutputDescriptor *vampGetOutputDescriptor(VampPluginHandle handle,
                                                         unsigned int i);

      static void vampReleaseOutputDescriptor(VampOutputDescriptor *desc);

      static VampFeatureList *vampProcess(VampPluginHandle handle,
                                          const float *const *inputBuffers,
                                          int sec,
                                          int nsec);

      static VampFeatureList *vampGetRemainingFeatures(VampPluginHandle handle);

      static void vampReleaseFeatureSet(VampFeatureList *fs);

      void checkOutputMap(plugin* p);
      void markOutputsChanged(plugin* p);

      void cleanup(plugin* p);
      unsigned int getOutputCount(plugin* p);
      VampOutputDescriptor *getOutputDescriptor(plugin* p, unsigned int i);
      VampFeatureList *process(plugin* p, const float *const *inputBuffers, int sec, int nsec);
      VampFeatureList *getRemainingFeatures(plugin* p);
      VampFeatureList *convertFeatures(plugin* p, const plugin::FeatureSet &features);
      
      // maps both plugins and descriptors to adapters
      typedef std::map<const void *, Impl *> AdapterMap;

      static AdapterMap *m_adapterMap;

      static std::mutex &adapterMapMutex() {
          static std::mutex m;
          return m;
      }
          
      static Impl *lookupAdapter(VampPluginHandle);

      std::mutex m_mutex; // guards all of the below
      
      bool m_populated;
      VampPluginDescriptor m_descriptor;
      plugin::parameter_descriptors m_parameters;
      plugin::programs m_programs;
      
      typedef std::map<plugin *, plugin::OutputList *> OutputMap;
      OutputMap m_pluginOutputs;

      std::map<plugin *, VampFeatureList *> m_fs;
      std::map<plugin *, std::vector<size_t> > m_fsizes;
      std::map<plugin *, std::vector<std::vector<size_t> > > m_fvsizes;
      void resizeFS(plugin* p, int n);
      void resizeFL(plugin* p, int n, size_t sz);
      void resizeFV(plugin* p, int n, int j, size_t sz);
  };

  PluginAdapterBase::PluginAdapterBase()
  {
      m_impl = new Impl(this);
  }

  PluginAdapterBase::~PluginAdapterBase()
  {
      delete m_impl;
  }

  const VampPluginDescriptor *
  PluginAdapterBase::getDescriptor()
  {
      return m_impl->getDescriptor();
  }

  PluginAdapterBase::Impl::Impl(PluginAdapterBase *base) :
      m_base(base),
      m_populated(false)
  {

      (void)adapterMapMutex(); // see comment in adapterMapMutex function above
  }

  const VampPluginDescriptor *
  PluginAdapterBase::Impl::getDescriptor()
  {

      std::lock_guard<std::mutex> guard(m_mutex);
      
      if (m_populated) return &m_descriptor;
      
      plugin *p = m_base->createPlugin(48000);
    
      if (!p) {
        std::cerr << "PluginAdapterBase::Impl::getDescriptor: Failed to create plugin" << std::endl;
          return 0;
      }

      m_parameters = p->get_parameter_descriptors();
      m_programs = p->get_programs();

      m_descriptor.identifier = strdup(p->get_identifier().c_str());
      m_descriptor.name = strdup(p->get_name().c_str());
      m_descriptor.description = strdup(p->get_description().c_str());
      m_descriptor.maker = strdup(p->get_maker().c_str());
      
      m_descriptor.parameterCount = m_parameters.size();
      m_descriptor.parameters = (const VampParameterDescriptor **)
          malloc(m_parameters.size() * sizeof(VampParameterDescriptor));

      unsigned int i;
      
      for (i = 0; i < m_parameters.size(); ++i) {
          VampParameterDescriptor *desc = (VampParameterDescriptor *)
              malloc(sizeof(VampParameterDescriptor));
          desc->identifier = strdup(m_parameters[i].identifier.c_str());
          desc->name = strdup(m_parameters[i].name.c_str());
          desc->description = strdup(m_parameters[i].description.c_str());
          desc->unit = strdup(m_parameters[i].unit.c_str());
          desc->minValue = m_parameters[i].min_value;
          desc->maxValue = m_parameters[i].max_value;
          desc->defaultValue = m_parameters[i].default_value;
          desc->isQuantized = m_parameters[i].is_quantized;
          desc->quantizeStep = m_parameters[i].quantize_step;
          desc->valueNames = 0;
          if (desc->isQuantized && !m_parameters[i].value_names.empty()) {
              desc->valueNames = (const char **)
                  malloc((m_parameters[i].value_names.size()+1) * sizeof(char *));
              for (unsigned int j = 0; j < m_parameters[i].value_names.size(); ++j) {
                  desc->valueNames[j] = strdup(m_parameters[i].value_names[j].c_str());
              }
              desc->valueNames[m_parameters[i].value_names.size()] = 0;
          }
          m_descriptor.parameters[i] = desc;
      }
      
      m_descriptor.programCount = m_programs.size();
      m_descriptor.programs = (const char **)
          malloc(m_programs.size() * sizeof(const char *));
      
      for (i = 0; i < m_programs.size(); ++i) {
          m_descriptor.programs[i] = strdup(m_programs[i].c_str());
      }
      
      if (p->getInputDomain() == plugin::FrequencyDomain) {
          m_descriptor.inputDomain = vampFrequencyDomain;
      } else {
          m_descriptor.inputDomain = vampTimeDomain;
      }

      m_descriptor.instantiate = vampInstantiate;
      m_descriptor.cleanup = vampCleanup;
      m_descriptor.initialise = vampInitialise;
      m_descriptor.reset = vampReset;
      m_descriptor.getParameter = vampGetParameter;
      m_descriptor.setParameter = vampSetParameter;
      m_descriptor.getCurrentProgram = vampGetCurrentProgram;
      m_descriptor.selectProgram = vampSelectProgram;
      m_descriptor.getPreferredStepSize = vampGetPreferredStepSize;
      m_descriptor.getPreferredBlockSize = vampGetPreferredBlockSize;
      m_descriptor.getMinChannelCount = vampGetMinChannelCount;
      m_descriptor.getMaxChannelCount = vampGetMaxChannelCount;
      m_descriptor.getOutputCount = vampGetOutputCount;
      m_descriptor.getOutputDescriptor = vampGetOutputDescriptor;
      m_descriptor.releaseOutputDescriptor = vampReleaseOutputDescriptor;
      m_descriptor.process = vampProcess;
      m_descriptor.getRemainingFeatures = vampGetRemainingFeatures;
      m_descriptor.releaseFeatureSet = vampReleaseFeatureSet;

      std::lock_guard<std::mutex> adapterMapGuard(adapterMapMutex());
      
      if (!m_adapterMap) {
          m_adapterMap = new AdapterMap;
      }
      (*m_adapterMap)[&m_descriptor] = this;

      delete p;

      m_populated = true;
      return &m_descriptor;
  }

  PluginAdapterBase::Impl::~Impl()
  {
      std::lock_guard<std::mutex> guard(m_mutex);

      if (!m_populated) return;

      free((void *)m_descriptor.identifier);
      free((void *)m_descriptor.name);
      free((void *)m_descriptor.description);
      free((void *)m_descriptor.maker);
      free((void *)m_descriptor.copyright);
          
      for (unsigned int i = 0; i < m_descriptor.parameterCount; ++i) {
          const VampParameterDescriptor *desc = m_descriptor.parameters[i];
          free((void *)desc->identifier);
          free((void *)desc->name);
          free((void *)desc->description);
          free((void *)desc->unit);
          if (desc->valueNames) {
              for (unsigned int j = 0; desc->valueNames[j]; ++j) {
                  free((void *)desc->valueNames[j]);
              }
              free((void *)desc->valueNames);
          }
          free((void *)desc);
      }
      free((void *)m_descriptor.parameters);

      for (unsigned int i = 0; i < m_descriptor.programCount; ++i) {
          free((void *)m_descriptor.programs[i]);
      }
      free((void *)m_descriptor.programs);

      std::lock_guard<std::mutex> adapterMapGuard(adapterMapMutex());
      
      if (m_adapterMap) {
          
          m_adapterMap->erase(&m_descriptor);
          
          if (m_adapterMap->empty()) {
              delete m_adapterMap;
              m_adapterMap = 0;
          }
      }
  }

  PluginAdapterBase::Impl *
  PluginAdapterBase::Impl::lookupAdapter(VampPluginHandle handle)
  {
      std::lock_guard<std::mutex> adapterMapGuard(adapterMapMutex());
      
      if (!m_adapterMap) return 0;
      AdapterMap::const_iterator i = m_adapterMap->find(handle);
      if (i == m_adapterMap->end()) return 0;
      return i->second;
  }

  VampPluginHandle
  PluginAdapterBase::Impl::vampInstantiate(const VampPluginDescriptor *desc,
                                           float inputSampleRate)
  {
    std::lock_guard<std::mutex> adapterMapGuard(adapterMapMutex());
      
      if (!m_adapterMap) {
          m_adapterMap = new AdapterMap();
      }

      if (m_adapterMap->find(desc) == m_adapterMap->end()) {
        std::cerr << "WARNING: PluginAdapterBase::Impl::vampInstantiate: Descriptor " << desc << " not in adapter map" << std::endl;
          return 0;
      }

      Impl *adapter = (*m_adapterMap)[desc];
      if (desc != &adapter->m_descriptor) return 0;

      plugin* p = adapter->m_base->createPlugin(inputSampleRate);
      if (p) {
          (*m_adapterMap)[p] = adapter;
      }
      return p;
  }

  void
  PluginAdapterBase::Impl::vampCleanup(VampPluginHandle handle)
  {

      Impl *adapter = lookupAdapter(handle);
      if (!adapter) {
          delete ((plugin*)handle);
          return;
      }
      adapter->cleanup(((plugin*)handle));
  }

  int
  PluginAdapterBase::Impl::vampInitialise(VampPluginHandle handle,
                                          unsigned int channels,
                                          unsigned int stepSize,
                                          unsigned int blockSize)
  {

      Impl *adapter = lookupAdapter(handle);
      if (!adapter) return 0;
      bool result = ((plugin*)handle)->initialise(channels, stepSize, blockSize);
      adapter->markOutputsChanged((plugin*)handle);
      return result ? 1 : 0;
  }

  void
  PluginAdapterBase::Impl::vampReset(VampPluginHandle handle) 
  {
      ((plugin *)handle)->reset();
  }

  float
  PluginAdapterBase::Impl::vampGetParameter(VampPluginHandle handle,
                                      int param) 
  {
      Impl *adapter = lookupAdapter(handle);
      if (!adapter) return 0.0;
      plugin::parameter_descriptors &list = adapter->m_parameters;
      return ((plugin *)handle)->get_parameter(list[param].identifier);
  }

  void
  PluginAdapterBase::Impl::vampSetParameter(VampPluginHandle handle,
                                      int param, float value)
  {
      Impl *adapter = lookupAdapter(handle);
      if (!adapter) return;
      plugin::parameter_descriptors &list = adapter->m_parameters;
      ((plugin *)handle)->set_parameter(list[param].identifier, value);
      adapter->markOutputsChanged((plugin *)handle);
  }

  unsigned int
  PluginAdapterBase::Impl::vampGetCurrentProgram(VampPluginHandle handle)
  {
      Impl *adapter = lookupAdapter(handle);
      if (!adapter) return 0;
      plugin::programs &list = adapter->m_programs;
      std::string program = ((plugin *)handle)->get_current_program();
      for (unsigned int i = 0; i < list.size(); ++i) {
          if (list[i] == program) return i;
      }
      return 0;
  }

  void
  PluginAdapterBase::Impl::vampSelectProgram(VampPluginHandle handle,
                                             unsigned int program)
  {
#ifdef DEBUG_PLUGIN_ADAPTER
      cerr << "PluginAdapterBase::Impl::vampSelectProgram(" << handle << ", " << program << ")" << endl;
#endif

      Impl *adapter = lookupAdapter(handle);
      if (!adapter) return;

      plugin::programs &list = adapter->m_programs;
      ((plugin *)handle)->select_program(list[program]);

      adapter->markOutputsChanged((plugin *)handle);
  }

  unsigned int
  PluginAdapterBase::Impl::vampGetPreferredStepSize(VampPluginHandle handle)
  {
      return ((plugin *)handle)->getPreferredStepSize();
  }

  unsigned int
  PluginAdapterBase::Impl::vampGetPreferredBlockSize(VampPluginHandle handle) 
  {
      return ((plugin *)handle)->getPreferredBlockSize();
  }

  unsigned int
  PluginAdapterBase::Impl::vampGetMinChannelCount(VampPluginHandle handle)
  {
      return ((plugin *)handle)->getMinChannelCount();
  }

  unsigned int PluginAdapterBase::Impl::vampGetMaxChannelCount(VampPluginHandle handle)
  {
      return ((plugin *)handle)->getMaxChannelCount();
  }

  unsigned int
  PluginAdapterBase::Impl::vampGetOutputCount(VampPluginHandle handle)
  {
      Impl *adapter = lookupAdapter(handle);

      if (!adapter) return 0;
      return adapter->getOutputCount((plugin *)handle);
  }

  VampOutputDescriptor *
  PluginAdapterBase::Impl::vampGetOutputDescriptor(VampPluginHandle handle,
                                                   unsigned int i)
  {
      Impl *adapter = lookupAdapter(handle);

      if (!adapter) return 0;
      return adapter->getOutputDescriptor((plugin *)handle, i);
  }

  void
  PluginAdapterBase::Impl::vampReleaseOutputDescriptor(VampOutputDescriptor *desc)
  {
      if (desc->identifier) free((void *)desc->identifier);
      if (desc->name) free((void *)desc->name);
      if (desc->description) free((void *)desc->description);
      if (desc->unit) free((void *)desc->unit);
      if (desc->hasFixedBinCount && desc->binNames) {
          for (unsigned int i = 0; i < desc->binCount; ++i) {
              if (desc->binNames[i]) {
                  free((void *)desc->binNames[i]);
              }
          }
      }
      if (desc->binNames) free((void *)desc->binNames);
      free((void *)desc);
  }

  VampFeatureList * PluginAdapterBase::Impl::vampProcess(VampPluginHandle handle,
                                       const float *const *inputBuffers,
                                       int sec,
                                       int nsec)
  {
      Impl *adapter = lookupAdapter(handle);
      if (!adapter) return 0;
      return adapter->process((plugin *)handle, inputBuffers, sec, nsec);
  }

  VampFeatureList *
  PluginAdapterBase::Impl::vampGetRemainingFeatures(VampPluginHandle handle)
  {
      Impl *adapter = lookupAdapter(handle);
      if (!adapter) return 0;
      return adapter->getRemainingFeatures((plugin *)handle);
  }

  void
  PluginAdapterBase::Impl::vampReleaseFeatureSet(VampFeatureList *)
  {
  }

  void 
  PluginAdapterBase::Impl::cleanup(plugin* p)
  {
      
    std::lock_guard<std::mutex> adapterMapGuard(adapterMapMutex());
    std::lock_guard<std::mutex> guard(m_mutex);
      
      if (m_fs.find(p) != m_fs.end()) {
          size_t outputCount = 0;
          if (m_pluginOutputs[p]) {
              outputCount = m_pluginOutputs[p]->size();
          }
          VampFeatureList *list = m_fs[p];
          for (unsigned int i = 0; i < outputCount; ++i) {
              for (unsigned int j = 0; j < m_fsizes[p][i]; ++j) {
                  if (list[i].features[j].v1.label) {
                      free(list[i].features[j].v1.label);
                  }
                  if (list[i].features[j].v1.values) {
                      free(list[i].features[j].v1.values);
                  }
              }
              if (list[i].features) free(list[i].features);
          }
          if (list) free((void *)list);
          m_fs.erase(p);
          m_fsizes.erase(p);
          m_fvsizes.erase(p);
      }

      if (m_pluginOutputs.find(p) != m_pluginOutputs.end()) {
          delete m_pluginOutputs[p];
          m_pluginOutputs.erase(p);
      }

      if (m_adapterMap) {
          m_adapterMap->erase(p);

          if (m_adapterMap->empty()) {
              delete m_adapterMap;
              m_adapterMap = 0;
          }
      }

      delete ((plugin *)p);
  }

  void 
  PluginAdapterBase::Impl::checkOutputMap(plugin* p)
  {
      // must be called with m_mutex held
      
      OutputMap::iterator i = m_pluginOutputs.find(p);

      if (i == m_pluginOutputs.end() || !i->second) {

          m_pluginOutputs[p] = new plugin::OutputList(p->getOutputDescriptors());
      }
  }

  void
  PluginAdapterBase::Impl::markOutputsChanged(plugin* p)
  {
    std::lock_guard<std::mutex> guard(m_mutex);

      OutputMap::iterator i = m_pluginOutputs.find(p);

      if (i != m_pluginOutputs.end()) {

          plugin::OutputList *list = i->second;
          m_pluginOutputs.erase(i);
          delete list;
      }
  }

  unsigned int 
  PluginAdapterBase::Impl::getOutputCount(plugin* p)
  {
    std::lock_guard<std::mutex> guard(m_mutex);

      checkOutputMap(p);

      return m_pluginOutputs[p]->size();
  }

  VampOutputDescriptor *
  PluginAdapterBase::Impl::getOutputDescriptor(plugin* p, unsigned int i)
  {
    std::lock_guard<std::mutex> guard(m_mutex);

      checkOutputMap(p);

      plugin::OutputDescriptor &od = (*m_pluginOutputs[p])[i];

      VampOutputDescriptor *desc = (VampOutputDescriptor *)
          malloc(sizeof(VampOutputDescriptor));

      desc->identifier = strdup(od.identifier.c_str());
      desc->name = strdup(od.name.c_str());
      desc->description = strdup(od.description.c_str());
      desc->unit = strdup(od.unit.c_str());
      desc->hasFixedBinCount = od.hasFixedBinCount;
      desc->binCount = od.binCount;

      if (od.hasFixedBinCount && od.binCount > 0) 
      {
          desc->binNames = (const char **)
              malloc(od.binCount * sizeof(const char *));
          
          for (unsigned int i = 0; i < od.binCount; ++i) {
              if (i < od.binNames.size()) {
                  desc->binNames[i] = strdup(od.binNames[i].c_str());
              } else {
                  desc->binNames[i] = 0;
              }
          }
      } else {
          desc->binNames = 0;
      }

      desc->hasKnownExtents = od.hasKnownExtents;
      desc->minValue = od.minValue;
      desc->maxValue = od.maxValue;
      desc->isQuantized = od.isQuantized;
      desc->quantizeStep = od.quantizeStep;

      switch (od.sampleType) {
      case plugin::OutputDescriptor::OneSamplePerStep:
          desc->sampleType = vampOneSamplePerStep; break;
      case plugin::OutputDescriptor::FixedSampleRate:
          desc->sampleType = vampFixedSampleRate; break;
      case plugin::OutputDescriptor::VariableSampleRate:
          desc->sampleType = vampVariableSampleRate; break;
      }

      desc->sampleRate = od.sampleRate;
      desc->hasDuration = od.hasDuration;

      return desc;
  }
      
  VampFeatureList *
  PluginAdapterBase::Impl::process(plugin *plugin,
                                   const float *const *inputBuffers,
                                   int sec, int nsec)
  {
  //    cerr << "PluginAdapterBase::Impl::process" << endl;

      real_time rt(sec, nsec);

      // We don't want to hold the mutex during the actual process call,
      // only while looking up the associated metadata
      {    
        std::lock_guard<std::mutex> guard(m_mutex);
          checkOutputMap(plugin);
      }

      return convertFeatures(plugin, plugin->process(inputBuffers, rt));
  }
      
  VampFeatureList *
  PluginAdapterBase::Impl::getRemainingFeatures(plugin* p)
  {
      {    
        std::lock_guard<std::mutex> guard(m_mutex);
          checkOutputMap(p);
      }

      return convertFeatures(p, p->getRemainingFeatures());
  }

  VampFeatureList *
  PluginAdapterBase::Impl::convertFeatures(plugin* p, const plugin::FeatureSet &features)
  {
    std::lock_guard<std::mutex> guard(m_mutex);

      int lastN = -1;

      int outputCount = 0;
      if (m_pluginOutputs[p]) outputCount = m_pluginOutputs[p]->size();
      
      resizeFS(p, outputCount);
      VampFeatureList *fs = m_fs[p];

  //    cerr << "PluginAdapter(v2)::convertFeatures: NOTE: sizeof(Feature) == " << sizeof(Plugin::Feature) << ", sizeof(VampFeature) == " << sizeof(VampFeature) << ", sizeof(VampFeatureList) == " << sizeof(VampFeatureList) << endl;

      for (plugin::FeatureSet::const_iterator fi = features.begin();
           fi != features.end(); ++fi) {

          int n = fi->first;
          
  //        cerr << "PluginAdapterBase::Impl::convertFeatures: n = " << n << endl;

          if (n >= int(outputCount)) {
            std::cerr << "WARNING: PluginAdapterBase::Impl::convertFeatures: Too many outputs from plugin (" << n+1 << ", only should be " << outputCount << ")" << std::endl;
              continue;
          }

          if (n > lastN + 1) {
              for (int i = lastN + 1; i < n; ++i) {
                  fs[i].featureCount = 0;
              }
          }

          const plugin::FeatureList &fl = fi->second;

          size_t sz = fl.size();
          if (sz > m_fsizes[p][n]) resizeFL(p, n, sz);
          fs[n].featureCount = sz;
          
          for (size_t j = 0; j < sz; ++j) {

  //            cerr << "PluginAdapterBase::Impl::convertFeatures: j = " << j << endl;

              VampFeature *feature = &fs[n].features[j].v1;

              feature->hasTimestamp = fl[j].hasTimestamp;
              feature->sec = fl[j].timestamp.sec;
              feature->nsec = fl[j].timestamp.nsec;
              feature->valueCount = fl[j].values.size();

              VampFeatureV2 *v2 = &fs[n].features[j + sz].v2;
              
              v2->hasDuration = fl[j].hasDuration;
              v2->durationSec = fl[j].duration.sec;
              v2->durationNsec = fl[j].duration.nsec;

              if (feature->label) free(feature->label);

              if (fl[j].label.empty()) {
                  feature->label = 0;
              } else {
                  feature->label = strdup(fl[j].label.c_str());
              }

              if (feature->valueCount > m_fvsizes[p][n][j]) {
                  resizeFV(p, n, j, feature->valueCount);
              }

              for (unsigned int k = 0; k < feature->valueCount; ++k) {
  //                cerr << "PluginAdapterBase::Impl::convertFeatures: k = " << k << endl;
                  feature->values[k] = fl[j].values[k];
              }
          }

          lastN = n;
      }

      if (lastN == -1) return 0;

      if (int(outputCount) > lastN + 1) {
          for (int i = lastN + 1; i < int(outputCount); ++i) {
              fs[i].featureCount = 0;
          }
      }

      return fs;
  }

  void
  PluginAdapterBase::Impl::resizeFS(plugin* p, int n)
  {

      int i = m_fsizes[p].size();
      if (i >= n) return;

      m_fs[p] = (VampFeatureList *)realloc
          (m_fs[p], n * sizeof(VampFeatureList));

      while (i < n) {
          m_fs[p][i].featureCount = 0;
          m_fs[p][i].features = 0;
          m_fsizes[p].push_back(0);
          m_fvsizes[p].push_back(std::vector<size_t>());
          i++;
      }
  }

  void
  PluginAdapterBase::Impl::resizeFL(plugin* p, int n, size_t sz)
  {
      size_t i = m_fsizes[p][n];
      if (i >= sz) return;

      m_fs[p][n].features = (VampFeatureUnion *)realloc
          (m_fs[p][n].features, 2 * sz * sizeof(VampFeatureUnion));

      while (m_fsizes[p][n] < sz) {
          m_fs[p][n].features[m_fsizes[p][n]].v1.hasTimestamp = 0;
          m_fs[p][n].features[m_fsizes[p][n]].v1.valueCount = 0;
          m_fs[p][n].features[m_fsizes[p][n]].v1.values = 0;
          m_fs[p][n].features[m_fsizes[p][n]].v1.label = 0;
          m_fs[p][n].features[m_fsizes[p][n] + sz].v2.hasDuration = 0;
          m_fvsizes[p][n].push_back(0);
          m_fsizes[p][n]++;
      }
  }

  void
  PluginAdapterBase::Impl::resizeFV(plugin* p, int n, int j, size_t sz)
  {
      
      size_t i = m_fvsizes[p][n][j];
      if (i >= sz) return;

      m_fs[p][n].features[j].v1.values = (float *)realloc
          (m_fs[p][n].features[j].v1.values, sz * sizeof(float));

      m_fvsizes[p][n][j] = sz;
  }
    
  PluginAdapterBase::Impl::AdapterMap *
  PluginAdapterBase::Impl::m_adapterMap = 0;
}
