#include "real_time.hpp"

#include <iostream>
#include <limits>
#include <sstream>
#include <sys/time.h>

namespace vamp 
{
  const real_time real_time::zero_time{ 0, 0 };

  real_time::real_time(int s, int n) : sec{ s }, nsec{ n }
  {
    while (nsec <= -ONE_BILLION && sec > std::numeric_limits<int>().max()) { nsec += ONE_BILLION; --sec; }
    while (nsec >=  ONE_BILLION && sec < std::numeric_limits<int>().max()) { nsec -= ONE_BILLION; ++sec; }
    while (nsec > 0 && sec < 0) { nsec -= ONE_BILLION; ++sec; }
    while (nsec < 0 && sec > 0) { nsec += ONE_BILLION; --sec; }
  }

  real_time real_time::from_seconds(double sec)
  {
    if (sec != sec) 
      return real_time::zero_time;
    else if (sec >= 0) 
      return real_time{ static_cast<int>(sec), static_cast<int>((sec - int(sec)) * ONE_BILLION + 0.5) };
    else 
      return -from_seconds(-sec);
  }

  real_time real_time::from_milliseconds(int msec)
  {
    return real_time{ msec / 1'000, (msec % 1'000) * 1'000'000 };
  }

  real_time real_time::from_timeval(const struct timeval& tv)
  {
    return real_time{ int(tv.tv_sec), int(tv.tv_usec * 1000) };
  }

  std::ostream& operator<<(std::ostream& out, const real_time& rt)
  {
    if (rt < real_time::zero_time) 
      out << "-";
    else 
      out << " ";
    
    int s = (rt.sec < 0 ? -rt.sec : rt.sec);
    int n = (rt.nsec < 0 ? -rt.nsec : rt.nsec);

    out << s << ".";

    int nn{ n };
    if (nn == 0) out << "00000000";
    else while (nn < (ONE_BILLION / 10)) 
    {
      out << "0";
      nn *= 10;
    }
    
    out << n << "R";
    return out;
  }

  std::string real_time::to_string() const
  {
    std::stringstream out;
    out << *this;
    std::string s = out.str();

    // remove trailing R
    return s.substr(0, s.length() - 1);
  }

  std::string real_time::to_text(bool fixed_dp) const
  {
    if (*this < real_time::zero_time) return "-" + (-*this).to_text(fixed_dp);

    std::stringstream out;

    if (sec >= 3'600) out << (sec / 3'600) << ":";
    
    if (sec >= 60) 
    {
      int minutes = (sec % 3'600) / 60;
      if (sec >= 3'600 && minutes < 10) out << "0";
      out << minutes << ":";
    }
    
    if (sec >= 10) out << ((sec % 60) / 10);
    
    out << (sec % 10);
    
    int ms = msec();

    if (ms != 0) 
    {
      out << ".";
      out << (ms / 100);
      ms = ms % 100;
      if (ms != 0) 
      {
        out << (ms / 10);
        ms = ms % 10;
      } 
      else if (fixed_dp) 
      {
        out << "0";
      }
      if (ms != 0) 
      {
        out << ms;
      } 
      else if (fixed_dp) 
      {
        out << "0";
      }
    } 
    else if (fixed_dp) 
    {
      out << ".000";
    }
  
    std::string s = out.str();

    return s;
  }

  real_time real_time::operator/(int d) const
  {
    int secdiv = sec / d;
    int secrem = sec % d;
    double nsecdiv = (static_cast<double>(nsec) + ONE_BILLION * static_cast<double>(secrem)) / d;
    return real_time{ secdiv, static_cast<int>(nsecdiv + 0.5) };
  }

  double real_time::operator/(const real_time& r) const
  {
    double l_total = static_cast<double>(sec) * ONE_BILLION + static_cast<double>(nsec);
    double r_total = static_cast<double>(r.sec) * ONE_BILLION + static_cast<double>(r.nsec);
    if (r_total == 0) return 0.0;
    else return l_total / r_total;
  }

  long real_time::rt2f(const real_time& time, unsigned int sample_rate)
  {
    if (time < real_time::zero_time) return -rt2f(-time, sample_rate);
    double s = time.sec + static_cast<double>(time.nsec) / ONE_BILLION;
    return static_cast<long>(s * sample_rate + 0.5);
  }

  real_time real_time::f2rt(long frame, unsigned int sample_rate)
  {
    if (frame < 0) return -f2rt(-frame, sample_rate);

    int sec = static_cast<int>(frame / static_cast<long>(sample_rate));
    frame -= sec * static_cast<long>(sample_rate);
    int nsec = static_cast<int>((static_cast<double>(frame) / static_cast<double>(sample_rate)) * ONE_BILLION + 0.5);
    return real_time{ sec, nsec };
  }
}
